function master_script(param)
% This function is the main function for the pipeline, and all other major functions are called by this function with different parameters. 
% Input: param: user defined parameter. 
%

%
% Author: Xiongtao Ruan
%
% Copyright (C) 2012-2017 Murphy Lab
% Computational Biology Department
% School of Computer Science
% Carnegie Mellon University
%
% This program is free software; you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published
% by the Free Software Foundation; either version 2 of the License,
% or (at your option) any later version.
%
% This program is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
% General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software
% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
% 02110-1301, USA.
%
% For additional information visit http://murphylab.web.cmu.edu or
% send email to murphy@cmu.edu





close all;
path(pathdef);
filename = mfilename;
base_filename = [filename, '/']; 

warning('off', 'MATLAB:MKDIR:DirectoryExists');
warning('off', 'Images:initSize:adjustingMag');
warning('off', 'MATLAB:xlsread:Mode');
warning('off', 'MATLAB:tifftagsread:rational:badTagValueDivisionByZero');

s = RandStream('mt19937ar', 'Seed', 905);
RandStream.setGlobalStream(s);   

default_params = master_script_get_default_params;

if ~exist('param', 'var')
    param = default_params;
else
    options_prior_to_processing = param;
end
param = master_script_process_options(param);

param
named_options_set = param.named_options_set;

t_cell_info = struct();
t_cell_info.param = param;

param.model_string = [param.causal_method, '_', param.opt_std];

results_location = param.results_location;
data_location  = [param.results_location, 'data/'];
clustering_result_location  = [param.results_location, 'clustering/'];
cross_validation_location  = [param.results_location, 'cross_validation/'];
model_location  = [param.results_location, 'models/'];
permutation_test_location  = [param.results_location, 'permutation_tests/'];
figure_location = [param.results_location, 'figures/'];

model_sub_location = [model_location, param.model_string, '/'];
figure_sub_location = [figure_location, param.model_string, '/'];
permutation_sub_location = [permutation_test_location, param.model_string, '/'];
permutation_computing_location = [permutation_sub_location, 'computing/'];

cv_sub_location = [cross_validation_location, param.model_string, '/'];
cv_data_location = [cross_validation_location, 'cross_validation_data/'];
cv_clustering_location = [cross_validation_location, 'cross_validation_clustering_data/'];
cv_computing_location = [cv_sub_location, 'computing/'];


mkdir(results_location);
mkdir(data_location);
mkdir(clustering_result_location);
mkdir(cross_validation_location);
mkdir(model_location);
mkdir(permutation_test_location);
mkdir(figure_location);
mkdir(model_sub_location);
mkdir(figure_sub_location);
mkdir(permutation_sub_location);
mkdir(permutation_computing_location);

mkdir(cv_sub_location);
mkdir(cv_data_location);
mkdir(cv_clustering_location);
mkdir(cv_computing_location);

t_cell_info.path_info = struct();
t_cell_info.path_info.tcell_models_location = param.tcell_models_location;
t_cell_info.path_info.results_location = results_location;
t_cell_info.path_info.data_location = data_location;
t_cell_info.path_info.clustering_result_location = clustering_result_location;
t_cell_info.path_info.cross_validation_location = cross_validation_location;
t_cell_info.path_info.model_location = model_location;
t_cell_info.path_info.permutation_test_location = permutation_test_location;
t_cell_info.path_info.figure_location = figure_location;

t_cell_info.path_info.model_sub_location = model_sub_location;
t_cell_info.path_info.figure_sub_location = figure_sub_location;
t_cell_info.path_info.permutation_sub_location = permutation_sub_location;
t_cell_info.path_info.permutation_computing_location = permutation_computing_location;

t_cell_info.path_info.cv_sub_location = cv_sub_location;
t_cell_info.path_info.cv_data_location = cv_data_location;
t_cell_info.path_info.cv_clustering_location = cv_clustering_location;
t_cell_info.path_info.cv_computing_location = cv_computing_location;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Add necessary code subdirectories to the path:

addpath(genpath('bonf_holm'))
addpath(genpath('pmkmp'))
addpath(genpath('export_fig-master'))
addpath(genpath('maximalCliques'))
addpath(genpath('kakearney-line2arrow-pkg-8aead6f'))
addpath(genpath('pmkmp'))

t_cell_info.param.illustration_colormap_to_use = pmkmp(1024, 'CubicL');

% download model data from the web

if ~param.skip_downoload_primary_data
    % don't need to use the script to download 
    % [t_cell_info] = master_script_download_raw_tcell_models(t_cell_info);
end

% get model data from t cell pipeline for models built from the raw images. 
[t_cell_info] = master_script_get_tcell_model_data(t_cell_info, param);    

% perform k-means clustering to form data
[t_cell_info] = master_script_clustering_Full_B7_data_extraction(t_cell_info, param);  

% hierachical clustering to show how to merge nodes, figure 2 in the paper. 
if false && strcmp(param.opt_std, 'CV_M2')
    [t_cell_info] = master_script_data_hierachical_clustering(t_cell_info);  
end

% cross validation to choose parameters
[t_cell_info] = master_script_cross_validation(t_cell_info, param);    

% build model
if ~strcmp(param.hostname, 'Xiongtao-PC-Office')
    switch param.causal_method  
        case 'CGP'
            [t_cell_info] = master_script_CGP_build_model(t_cell_info, param);
        case 'CENR'
            [t_cell_info] = master_script_CENR_build_model(t_cell_info, param);
    end
end

% postprocess scripts. 
if ~false 
% visualize adjacency matrices, figure 3 and figure 4 in the paper
[t_cell_info] = master_script_visualize_model(t_cell_info);

% visualize kinetics for cause and effect relationships, figure 5 in the paper
[t_cell_info] = master_script_plot_kinetics(t_cell_info);

% visualize spatiotemporal network, figure 6 in the paper
[t_cell_info] = master_script_plot_spatiotemporal_network(t_cell_info);

% visualize network comparison for two different methods, figure 7 in the paper
[t_cell_info] = master_script_plot_adjacency_network_comparison(t_cell_info);

% plot ROC curve for network reconstruction
[t_cell_info] = master_script_plot_network_ROC_curve(t_cell_info);

% Generate table 1 in the paper
[t_cell_info] = master_script_generate_error_comparison_table(t_cell_info);

end

% [t_cell_info] = master_script_plot_adjacency_network(t_cell_info);

% permutation tests, figure 8 in the paper 
if ~param.skip_permutation_tests
    [t_cell_info] = master_script_B7_permutation_testing(t_cell_info);
    [t_cell_info] = master_script_B7_permutation_testing_postprocess(t_cell_info);
end


end
