function [R_t_3d, A, c, W_err, W, W_r] = master_script_causal_graph_process(X, M, param)
% input:
% X: f X t, time series data
% M: lag time
% param: some parameters for optimization.
%
% Output:
% R_t_3d: f X f X M, pilynomial coeficients for X.
% A: f X f, adjacency matrix
% c: coefficients for P(A).
% W_err: residuals of the computing. 
% 
% this function is for solving A and c in causal graph process
% 
% Reference:
%    1. Mei, Jonathan, and José MF Moura. "Signal Processing on Graphs: Modeling 
%    (Causal) Relations in Big Data." arXiv preprint arXiv:1503.00173 (2015).
%    2. Figueiredo, Mário AT, Robert D. Nowak, and Stephen J. Wright. "Gradient 
%    projection for sparse reconstruction: Application to compressed sensing 
%    and other inverse problems." Selected Topics in Signal Processing, IEEE 
%    Journal of 1.4 (2007): 586-597.
%
%
% Author: Xiongtao Ruan xruan@andrew.cmu.edu
% Date: Nov. 16, 2015
%
% Jan. 27, 2016 Change initiation method. Set random number generator so
% that they can start from the same random point
% Jan. 29, 2016 Add the computing for residuals. 
% May. 3, 2016 formation of parameter structure. 
%

%
% Author: Xiongtao Ruan
%
% Copyright (C) 2012-2017 Murphy Lab
% Computational Biology Department
% School of Computer Science
% Carnegie Mellon University
%
% This program is free software; you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published
% by the Free Software Foundation; either version 2 of the License,
% or (at your option) any later version.
%
% This program is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
% General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software
% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
% 02110-1301, USA.
%
% For additional information visit http://murphylab.web.cmu.edu or
% send email to murphy@cmu.edu




% Add random number generator setting. 
s = RandStream('mt19937ar', 'Seed', 21335);
RandStream.setGlobalStream(s);

if nargin < 3
    param = struct();
end
if nargin < 2
    M = 1;
end
if nargin < 1
    N = 18;
    T = 12;
    flag = 3;
    [X, A_M_3d, Omg, A_M, c_M] = RandDataGenerator_graphs(M, N, T, flag);
    X = X(:, 1 : end);
    param.self_prediction = false;
    param.zero_mat = kron(eye(6), ones(3));
end

param = process_options_structure( struct('lambda_1', 0.0022, ...
                                          'lambda_2', 0.0001, ...
                                          'lambda_3', 0.001, ...
                                          'nonnegative', false, ...
                                          'optimize_c_method', 1, ...
                                          'self_prediction', true, ...
                                          'zero_mat', zeros(size(X, 1)), ...
                                          'reconstrution_error_method', 2 ...
                                      ), param);
                                  
Err = 10;

[m, n] = size(X);
nIter = 20;

% optimize R
% t0 = cputime
tic
[R_t] = optimize_R(X, M, param);
toc
% R_t
% t1 = cputime - t0
% save('R_t.mat', 'R_t');
R_t_3d = reshape(R_t, [m, m, M]);

% optimize A
% param.lambda_1 = 0.01;
% param.lambda_2 = 0.01;
param.nIter = 30;
% Jan. 27, 2016 for A_init initiation from the same point
param.A_init = squeeze(R_t_3d(:, :, 1));
param.A_init = rand(size(R_t_3d(:, :, 1)));
param.Err = 1e-7 * numel(param.A_init);
param.nonnegative = false;
% param.nonnegative = true;
A = optimize_A(R_t_3d, param);

% optimize c
% param.lambda_3 = 0.01;

param.nIter = 100;
param.c_init = zeros(M, M + 1);
optimize_c_method = param.optimize_c_method;
param.nonnegative = false;
c = zeros(M, M + 1);

switch optimize_c_method
    case 1
        for i = 1 : M 
            param.Err = 1e-6 * ( i + 1);
            param.ci_init = zeros(i + 1, 1);
            R_i = R_t_3d(:, :, i);
            ci = optimize_ci(i, A, R_i, param);
            c(i, 1 : i + 1) = ci';
        end
    case 2
        c_init = triu(transpose(param.c_init), -1);
        param.c_init = c_init(triu(ones(size(c_init)), -1) > 0);
        vec_c = optimize_c(X, A, M, param);
        tr_c = zeros(size(c'));
        tr_c(triu(ones(size(c')), -1) > 0) = vec_c;
        c = tr_c';
end

% computing of residuals 
% [W_err, W, W_r] = CGP_residual_computing(R_t_3d, A, c, M, X, param.reconstrution_error_method);


end

function [R] = optimize_R(X, M, param)

Err = 1e-5;
nIter = 100;
min_err = 1e20;
[m, n] = size(X);
% random start should be much better than initialization from zero
R_t = zeros(m, m * M);
% 05/11/2016
% R_t = 2 * rand(m, m * M) - 1;
R_tp1 = zeros(m, m * M);

% lambda_1 = 0.01;
% lambda_2 = 0.02;
% nonnegative = false;

% param.lambda_1 = lambda_1;
% param.lambda_2 = lambda_2;
param.nIter = 100;
% param.nonnegative = nonnegative;
param.M = M;


for i_iter = 1 : nIter
    R_t_3d = reshape(R_t, m, m, M);
    % R_tp1 = reshape(R_tp1, m, n, M);
    for i = 1 : M
        % R_i = squeeze(R_t_3d(:, :, i));
        param.i = i;
        if i == 1 
            param.nonnegative = false;
        end
        %tic
        R_i_est = optimize_R_i(X, R_t_3d, param);
        %toc
        R_t_3d(:, :, i) = R_i_est;
             
    end
    R_tp1 = reshape(R_t_3d, m, []);
    R_err = norm((R_tp1 - R_t), 'fro');
    
    if min_err > R_err
        min_err_old = min_err;
		min_err = R_err;
		min_R = R_tp1;
    end
    if min_err > Err
        R_t = R_tp1;
    else
        break;
    end
    if abs((min_err_old - min_err)) / min_err < 0.01
        break;
    end
     
end
R_tp1 = min_R;
R = R_tp1;
end


function [R_i] = optimize_R_i(X, R_t_3d, param)

beta = 0.5;
mu = 0.5;
alpha_0 = 1;
alpha_min = 1e-30;
if nargin < 3
    nIter = 30;
    lambda_1 = 0.01;
    lambda_2 = 0.02;
    nonnegative = false;
    i = 1;
    M = 3;
else
    lambda_1 = param.lambda_1;
    lambda_2 = param.lambda_2;
    % Err = param.Err;
    nIter = param.nIter;
    nonnegative = param.nonnegative;
    i = param.i;
    M = param.M;
end

% if i == 1
% 	Err = 30;
% else 
% 	Err = 3;
% end

R_i = squeeze(R_t_3d(:, :, i));
[m, n] = size(R_i);

Err = 1e-6 * numel(R_i);
z_k = [ R_i.* (R_i > 0), -R_i.* (R_i < 0)];
z_kp1 = 100 * ones(size(z_k));
min_err = 1e20;
for k = 1 : nIter   
    [F_zk, grad_F_zk] = FR_function_gradient_compute(X, R_t_3d, i, z_k, M, lambda_1, lambda_2, 2);
    alpha_k = alpha_0;
    w_k = (z_k - alpha_k * grad_F_zk);
    w_k(w_k < 0) = 0;
    if nonnegative 
        w_k(:, n + 1 : end) = 0;
    end
    F_wk = FR_function_gradient_compute(X, R_t_3d, i, w_k, M, lambda_1, lambda_2, 1);
    while F_wk > F_zk - mu * alpha_k * grad_F_zk(:)' * reshape( z_k - w_k, [], 1)
        alpha_k = beta * alpha_k;
        if alpha_k < alpha_min
            alpha_k = alpha_min;
            break;
        end
        w_k = (z_k - alpha_k * grad_F_zk);
        w_k(w_k < 0) = 0;
        if nonnegative 
            w_k(:, n + 1 : end) = 0;
        end
        if i == 1 && ~param.self_prediction
            w_k([param.zero_mat, param.zero_mat] == 1) = 0;
        end
        F_wk = FR_function_gradient_compute(X, R_t_3d, i, w_k, M, lambda_1, lambda_2, 1);
        % F_wk, F_zk - mu * grad_F_zk(:)' * reshape( z_k - w_k, [], 1)
    end
    
    % F_wk
    z_kp1 = w_k;
    current_err = norm(z_kp1 - z_k, 'fro');
    cur_F_err = abs(F_wk - F_zk);
    if current_err > Err || cur_F_err > Err
        z_k = z_kp1;
    else
        break;
    end
%     if abs((min_err_old - min_err)) / min_err < 0.01
%         break;
%     end
    
end

R_i = z_kp1(:, 1 : n) - z_kp1(:, n + 1 : end);

end


function [F, grad_F] = FR_function_gradient_compute(X, R_t_3d, i, z, M, lambda_1, lambda_2, flag)

b = zeros(size(X, 1));
K = size(X, 2);

u = z(:, 1 : size(z, 2) / 2);
v = z(:, size(z, 2) / 2 + 1 : end);

Y = X(:, M + 1 : end);
X_i = X(:, M - i + 1 : end - i);
b = Y * X_i';
for j = 1 : M 
    if j == i
        continue;
    end
    b = b - R_t_3d(:, :, j) * X(:, M - j + 1 : end - j) * X_i';
end

c = lambda_1 * ones(size(z)) + [-b, b];

B1 = zeros(size(X, 1));
B1_1 = X_i * X_i';
B1_2 = zeros(size(X, 1));

for j = 1 : M
    if j == i
        continue;
    end   
    B1_2 = B1_2 + R_t_3d(:, :, j) * R_t_3d(:, :, j)';
end

B1 = B1_1 + 2 * lambda_2 * B1_2;
% xruan 11/22/2015
% kron(A, B) is different from kron(B, A)
B1_kr = kron([1, -1; -1, 1], B1);
B3_kr = 2 * lambda_2 * kron( [1, -1; -1, 1], eye(size(X, 1)));
R_i = R_t_3d(:, :, i);

% z = [ R_i.* (R_i > 0), R_i.* (R_i < 0)];

F = z * c' + 0.5 * z * B1_kr * z';
% F = z * c' + 0.5 * (u * B1 * u' - u * B1 * v' - v * B1 * u' + v * B1 * v');
if flag == 1
    grad_F = zeros(size(c));
else
    % xruan
    % Dec 7, 2015
    % fixed the bug of gradient computing. 
    grad_F = c + 0.5 * z * (B1_kr + B1_kr');
    % zB1 = u * B1 - v * B1;
    % grad_F = c + [zB1, -zB1];
end

zB3zt = z * B3_kr * z';
zB3 = z * B3_kr;
Rjt_Rj = zeros(size(R_t_3d, 1));

for j = 1 : M
    if j == i
        continue;
    end
    R_j = R_t_3d(:, :, j);
    B2_kr_j =-4 * lambda_2 *  kron([1, -1; -1, 1], R_j');
    F = F + 0.5 * R_j * z * B2_kr_j * z';
    Rjt_Rj = Rjt_Rj + R_j' * R_j;
    if flag == 2
        grad_F = grad_F + 0.5 * ( R_j' * z * B2_kr_j' + R_j * z * B2_kr_j);
    end
end
F = F + 0.5 * Rjt_Rj * zB3zt;
F = trace(F);

if flag == 2
    grad_F = grad_F + Rjt_Rj * zB3;
end


end


function [A] = optimize_A(R_t_3d, param)

beta = 0.5;
mu = 0.5;
alpha_0 = 1;
alpha_min = 1e-30;

if nargin < 2
    lambda_1 = 0.01;
    lambda_2 = 0.02;
    Err = 1e-4;
    nIter = 30;
    A_init = R_t_3d(:, :, 1);
    nonnegative = true;
else
    lambda_1 = param.lambda_1;
    lambda_2 = param.lambda_2;
    Err = param.Err;
    nIter = param.nIter;
    A_init = param.A_init;
    nonnegative = param.nonnegative;
end

A = A_init;
z_k = [A .* (A > 0), -A.* (A < 0)];
z_kp1 = 100 * ones(size(z_k));
min_err = 1e20;
for k = 1 : nIter 
    
    [F_zk, grad_F_zk] = FA_function_gradient_compute(R_t_3d, z_k, lambda_1, lambda_2, 2);
    alpha_k = alpha_0;
    w_k = (z_k - alpha_k * grad_F_zk);
    w_k(w_k < 0) = 0;
    if nonnegative 
        w_k(:, size(A, 2) + 1 : end) = 0;
    end    
    F_wk = FA_function_gradient_compute(R_t_3d, w_k, lambda_1, lambda_2, 1);
    while F_wk > F_zk - mu * alpha_k * grad_F_zk(:)' * reshape( z_k - w_k, [], 1)
        alpha_k = beta * alpha_k;
        if alpha_k < alpha_min
            alpha_k = alpha_min;
            break;
        end
        w_k = (z_k - alpha_k * grad_F_zk);
        w_k(w_k < 0) = 0;
        if nonnegative 
            w_k(:, size(A, 2) + 1 : end) = 0;
        end
        if ~param.self_prediction
            w_k([param.zero_mat, param.zero_mat] == 1) = 0;
        end
        F_wk = FA_function_gradient_compute(R_t_3d, w_k, lambda_1, lambda_2, 1);
        % F_wk, F_zk - mu * grad_F_zk(:)' * reshape( z_k - w_k, [], 1)
    end
    
    % F_wk
    z_kp1 = w_k;
    current_err = norm(z_kp1 - z_k, 'fro');
	if min_err > current_err 
		min_err = current_err;
		min_z = z_kp1;
	end
    if min_err > Err || abs(F_wk - F_zk) > Err
        z_k = z_kp1;
    else
        break;
    end
end
z_kp1 = min_z;
A = z_kp1(:, 1 : size(A, 2)) - z_kp1(:, size(A, 2) + 1 : end);

end


function [F, grad_F] = FA_function_gradient_compute(R_t_3d, z, lambda_1, lambda_2, flag)
% compute F_A(z) and the gradient of F_A(z)

n = size(R_t_3d, 1);
M = size(R_t_3d, 3);

b = R_t_3d(:, :, 1);


c = lambda_1 * ones(size(z)) + [-b, b];

B1 = eye(n);

for j = 2 : M
    
    B1 = B1 + 2 * lambda_2 * R_t_3d(:, :, j) * R_t_3d(:, :, j)';
end

B1_kr = kron([1, -1; -1, 1], B1);
B3_kr = 2 * lambda_2 * kron([1, -1; -1, 1], eye(n));


F = z * c' + 0.5 * z * B1_kr * z';
% F = z * c' + 0.5 * (u * B1 * u' - u * B1 * v' - v * B1 * u' + v * B1 * v');
if flag == 1
    grad_F = zeros(size(c));
else
    % xruan Dec. 7, 2015
    grad_F = c + 0.5 * z * (B1_kr + B1_kr');
end

Rjt_Rj = zeros(size(R_t_3d, 1));

for j = 2 : M
    R_j = R_t_3d(:, :, j);
    B2_kr_j = -4 * lambda_2 *  kron([1, -1; -1, 1], R_j);

    F = F + 0.5 * R_j * z * B2_kr_j * z';
    Rjt_Rj = Rjt_Rj + R_j' * R_j;
    if flag == 2
        grad_F = grad_F + 0.5 * ( R_j' * z * B2_kr_j' + R_j * z * B2_kr_j);
    end
end

F = F + 0.5 * Rjt_Rj * z * B3_kr * z';
F = trace(F);

if flag == 2
    grad_F = grad_F + Rjt_Rj * z * B3_kr;
end


end

function [ci] = optimize_ci(i, A, R_i, param)
% use method 1 to estimate c

beta = 0.5;
mu = 0.5;
alpha_0 = 1;
alpha_min = 1e-30;

if nargin < 4
    lambda_3 = 0.01;
    Err = 1e-4 * (i + 1);
    nIter = 30;
    ci_init = zeros(i + 1, 1);
    nonnegative = true;
else
    lambda_3 = 0.01;
    Err = param.Err;
    nIter = param.nIter;
    ci_init = param.ci_init;
    nonnegative = param.nonnegative;

end

ci = ci_init;
z_k = [ci .* (ci > 0); ci.* (ci < 0)];
z_kp1 = 100 * ones(size(z_k));
min_err = 1e20;

n = size(A, 1);
% function handle of F_zk, grad_F_zk
Q_i = zeros(n ^ 2, i + 1);
% change to eye
% A_0 = ones(n);
A_0 = eye(n);
Q_i(:, 1) = A_0(:);
for j = 2 : i + 1
    A_j = A_0 * A;
    Q_i(:, j) = A_j(:);
end
b = Q_i' * R_i(:);
q = lambda_3 + [-b; b];
B = kron([1, -1; 1, -1], Q_i' * Q_i);
Fci_function = @(z) q' * z + 0.5 * z' * B * z;
Fci_gradient = @(z) q + B * z;

for k = 1 : nIter     
    F_zk = Fci_function(z_k);
    grad_F_zk = Fci_gradient(z_k);
    alpha_k = alpha_0;
    w_k = (z_k - alpha_k * grad_F_zk);
    w_k(w_k < 0) = 0;
    if nonnegative 
        w_k(size(ci, 1) + 1 : end) = 0;
    end    
    F_wk = Fci_function(w_k);
    while F_wk > F_zk - mu * alpha_k * grad_F_zk(:)' * reshape( z_k - w_k, [], 1)
        alpha_k = beta * alpha_k;
        if alpha_k < alpha_min;
            alpha_k = alpha_min;
            break;
        end
        w_k = (z_k - alpha_k * grad_F_zk);
        w_k(w_k < 0) = 0;
        if nonnegative 
            w_k(size(ci, 1) + 1 : end) = 0;
        end
        F_wk = Fci_function(w_k);
        % F_wk, F_zk - mu * grad_F_zk(:)' * reshape( z_k - w_k, [], 1)
    end
    
    z_kp1 = w_k;
    current_err = norm(z_kp1 - z_k, 'fro');
	if min_err > current_err
		min_err = current_err;
		min_z = z_kp1;
	end
    if min_err > Err
        z_k = z_kp1;
    else
        break;
    end
end
z_kp1 = min_z;
ci = z_kp1(1 : size(ci, 1)) - z_kp1(size(ci, 1) + 1 : end);

end

function [c] = optimize_c(X, A, M, param)
% use method 2 to optimize c

beta = 0.5;
mu = 0.5;
alpha_0 = 1;
alpha_min = 1e-30;

N_c = M * (M + 3) / 2;
if nargin < 4
    lambda_3 = 0.01;
    Err = 1e-4 * N_c;  % number of parameter is M(M+3)/2
    nIter = 30;
    c_init = zeros(N_c, 1);
    nonnegative = true;
else
    lambda_3 = 0.01;
    Err = param.Err;
    nIter = param.nIter;
    c_init = param.c_init;
    nonnegative = param.nonnegative;

end

c = c_init;
z_k = [c .* (c > 0); c.* (c < 0)];
z_kp1 = 100 * ones(size(z_k));
min_err = 1e20;

[n, T] = size(X);
% compute B and Y
Y = X(:, M + 1 : end);
Y = Y(:);
% B = zeros((T - M) * n, (M + 3) * M / 2);

A_02M = arrayfun(@(j) A ^ j, 0 : M, 'Uniformoutput', false);
B = [];
for i = 1 : M
    X_i = X(:, M + 1 - i : T - i);
    Aj_X_i = cellfun(@(Aj) Aj * X_i, A_02M(1 : i + 1), 'Uniformoutput', false);
    AjXi = cell2mat(cellfun(@(H) H(:), Aj_X_i, 'uniformoutput', false));
    B = [B, AjXi];
end

q = [-B' * Y; B' * Y] + lambda_3;
H = kron([1, -1; -1, 1], B' * B);

Fc_function = @(z) q' * z + 0.5 * z' * H * z;
Fc_gradient = @(z) q + H * z;

for k = 1 : nIter     
    F_zk = Fc_function(z_k);
    grad_F_zk = Fc_gradient(z_k);
    alpha_k = alpha_0;
    w_k = (z_k - alpha_k * grad_F_zk);
    w_k(w_k < 0) = 0;
    if nonnegative 
        w_k(size(c, 1) + 1 : end) = 0;
    end    
    F_wk = Fc_function(w_k);
    while F_wk > F_zk - mu * alpha_k * grad_F_zk(:)' * reshape( z_k - w_k, [], 1)
        alpha_k = beta * alpha_k;
        if alpha_k < alpha_min;
            alpha_k = alpha_min;
            break;
        end
        w_k = (z_k - alpha_k * grad_F_zk);
        w_k(w_k < 0) = 0;
        if nonnegative 
            w_k(size(c, 1) + 1 : end) = 0;
        end
        F_wk = Fc_function(w_k);
        % F_wk, F_zk - mu * grad_F_zk(:)' * reshape( z_k - w_k, [], 1)
    end
    
    z_kp1 = w_k;
    current_err = norm(z_kp1 - z_k, 'fro');
	if min_err > current_err
		min_err = current_err;
		min_z = z_kp1;
	end
    if min_err > Err
        z_k = z_kp1;
    else
        break;
    end
end
z_kp1 = min_z;
c = z_kp1(1 : size(c, 1)) - z_kp1(size(c, 1) + 1 : end);

end
