function [R_t_3d, A, c, f_mat, f_terms] = master_script_spatiotemporal_constraint_projected_admm(X, M, D, b, param)
% performing optimization to build model for CENR method. 
%
% input:
% X: time series data
% M: lag time
% D: preselected pairs represented in a adjacency matrix with binary
% entries. 
% b: strength of the constrains. 
% param: some parameters for optimization. 
%
% Output:
% R_t_3d: f X f X M, pilynomial coeficients for X.
% A: f X f, adjacency matrix
% c: coefficients for P(A).
% f_mat: objective functions for each iteration. 
% f_terms: different terms in the objective functions. 

%
% Author: Xiongtao Ruan
%
% Copyright (C) 2012-2017 Murphy Lab
% Computational Biology Department
% School of Computer Science
% Carnegie Mellon University
%
% This program is free software; you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published
% by the Free Software Foundation; either version 2 of the License,
% or (at your option) any later version.
%
% This program is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
% General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software
% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
% 02110-1301, USA.
%
% For additional information visit http://murphylab.web.cmu.edu or
% send email to murphy@cmu.edu






debug = false;
if nargin < 2
    M = 2;
end
if nargin < 1
    s = RandStream('mt19937ar', 'Seed', 21335);
    RandStream.setGlobalStream(s);   
    N = 36;
    T = 20;
    flag = 3;
    [X, A_M_3d, Omg, A_M, c_M] = RandDataGenerator_graphs(M, N, T, flag);
    X = X(:, 1 : end);
    debug = true;
    if ~false
        D = zeros(size(A_M_3d));
        b = 0;
        % D(1 : m, 1 : m) = ((rand(size(beta_0(1 : m, 1 : m))) > 0.5) - 0.5) * 2;
        % D(22) = 1;
        % D(6) = -1;
        % D(30) = -1;
        % D(6, 2) = -1;
        % D(7, 2) = -1;
        % D(18, 2) = -1;
        % D(5, 5) = 1;
        % D(27, 4) = 1;
        D1 = zeros(N);
        D1(A_M > 0.25) = -1;
        D1(A_M < -0.25) = 1;
        % D1(rand(m) < 0.9) = 0;
        D(1 : N, 1 : N) = D1;
        D = reshape(D, size(D, 1), []);
        b = -0.3;
    end 
end

if nargin < 3
    param = struct();
end

param = process_options_structure( struct('lambda_1', 0.01, ...
                                          'lambda_2', 0.01, ...
                                          'lambda_3', 0.001, ...
                                          'region_num', 3, ...
                                          'nonnegative', false, ...
                                          'learn_constant_term', false, ...
                                          'optimize_c_method', 1, ...
                                          'reconstrution_error_method', 1, ...
                                          'group_method', 2, ...
                                          'group_weight', 'square_root',...
                                          'self_prediction', true ...
                                      ), param);



lambdas = [param.lambda_1, param.lambda_2];
region_num = param.region_num;
learn_constant_term = param.learn_constant_term;
self_prediction = param.self_prediction;

X_0 = [];
for i = 1 : M
    X_0 = [X_0; X(:, M - i + 1 : end - i)];
end

X_1 = X(:, M + 1 : end);

[m, n] = size(X);

if learn_constant_term
    beta_0 = rand(m * (m +1), 1) - 0.5;
    X_0 = [X_0; ones(1, size(X_0, 2))];
else
    beta_0 = rand(m, m * M) - 0.5;
end

% X_vec = sparse(kron(X_0, eye(m)))';

rho = 1;
epsilon = 1e-9;
nIter = 5000;
Y = X_1;

[beta_k, f_mat, f_terms] = admm_group_lasso(Y, X_0, beta_0, D, b, rho, lambdas, epsilon, nIter, self_prediction);     


R_t_3d = reshape(beta_k, m, m, M);
A = beta_k(1 : m, 1 : m);
c = [zeros(M, 1), eye(M)];
if learn_constant_term
    A = A(:, 1 : end - 1);
end    

if ~false && debug
    % close all
    figure, 
    plot(f_mat);

    figure, 
    imagesc(reshape(beta_k, m, []));
    colorbar
    
    figure, 
    imagesc(reshape(beta_k(1 : m, 1 : m), m, []));
    title('A_1')
    colorbar
    
    figure, 
    imagesc(A_M);
    title('ground truth')
    colorbar
    norm(beta_k - reshape(A_M_3d, size(A_M_3d, 1), []), 'fro')  
end  
    
end

function [beta_k, f_mat, f_terms] = admm_group_lasso(y, X, beta_0, D, b, rho, lambdas, epsilon, nIter, self_prediction)

lambda_1 = lambdas(1);
lambda_2 = lambdas(2);

beta_k = beta_0;
Z_1 = beta_0;

s = max(beta_0, 0);
V = 0 * beta_0;
U_1 = 0 * beta_0;
N = size(beta_0, 1);

f_mat = zeros(nIter + 1, 1);

f_0 = compute_objective_function(y, X, beta_0, lambdas);
f_mat(1) = f_0;

Beta_project_func = @(beta, D, b) (D < 0) .* max(beta, -b) + (D > 0) .* min(beta, b) + (D == 0) .* beta;
Sign_func = @(z, t) (abs(z) > t) .* ( z - t * (z > t) + t * (z < - t));
beta_update_denominator = X * X' + (rho + 2 * lambda_2) * eye(size(X, 1));
yX_T = y * X';

for k = 1 : nIter       
    % beta_k = (X' * X + rho * eye(size(X, 2))) \ (X' * y + rho * (z1_k - w_k));
    % projected update for beta
    beta_k = Beta_project_func((yX_T + rho * (Z_1 - U_1)) / beta_update_denominator, D, b);
    if ~self_prediction
        beta_k(kron(ones(1, size(beta_k, 2) / size(beta_k, 1)), eye(size(beta_k, 1))) == 1) = 0;
    end
    % beta_k(abs(beta_k) < 1e-5) = 0;
    Z_1 = Sign_func(beta_k + U_1, lambda_1 / rho);
    U_1 = U_1 + beta_k - Z_1;
    
    f_ip1 = compute_objective_function(y, X, beta_k, lambdas);
    f_mat(k + 1) = f_ip1;
    if abs(f_mat(k + 1) - f_mat(k)) < epsilon && k < nIter
        f_mat(k + 2 : end) = [];
        break;
    end
end


if ~false
    f_terms = zeros(4, 1);
    
    lambda_1 = lambdas(1);
    lambda_2 = lambdas(2);
    f_terms(2) = 1 / 2 * norm(y - beta_k * X, 'fro') .^ 2;  
    f_terms(3) = lambda_1 * sum(abs(beta_k(:)));
    f_terms(4) = lambda_2 * norm(beta_k, 'fro') ^ 2;
    f_terms(1) = sum(f_terms);
end
end

function [z] = group_lasso_minimize_z_2(x, t_vec, group_labels_inds)

z = zeros(size(x));
for j = 1 : numel(group_labels_inds)
    cur_inds =  group_labels_inds{j};
    z(cur_inds) = prox_function_j(x(cur_inds), t_vec(j));
end

end

function [z] = prox_function_j(x, t)

if norm(x) <= t
    z = zeros(size(x));
else
    z = (1 - t / norm(x)) * x;
end
end

function [f] = compute_objective_function(y, X, beta, lambdas, W, group_labels_inds)

lambda_1 = lambdas(1);
lambda_2 = lambdas(2);
reg_term = lambda_1 * sum(abs(beta(:))) + lambda_2 * norm(beta, 'fro') ^ 2;
% reg_term = reg_term + sum(arrayfun(@(j) lambda_2 * W(j) * norm(beta(group_labels_inds{j}), 'fro'), 1 : numel(group_labels_inds)));
f = 1 / 2 * norm(y - beta * X, 'fro') .^ 2 + reg_term;

end
